import pyfaidx
import ganz

genome = pyfaidx.Fasta("/Users/danila/pd/hematopoiesis/data/ref/GRCh37 genome.fa")
sequence = {f'chr{i}': genome[f'chr{i}'] for i in range(1, 23)}
for chrom in sequence:
    sequence[chrom] = ganz.DnaSequence(sequence[chrom][:].seq)
